/*
    AQTRONIX C++ Library
    Copyright 2012 Parcifal Aertssen

    This file is part of AQTRONIX C++ Library.

    AQTRONIX C++ Library is free software; you can redistribute it
	and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
	the License, or (at your option) any later version.

    AQTRONIX C++ Library is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AQTRONIX C++ Library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
    MA  02111-1307  USA
*/
// SyslogSocket.h: interface for the SyslogSocket class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SyslogSocket.h"

// CSyslogSocket
CSyslogSocket::CSyslogSocket():m_IsReady(false),LastError(""),SourceHostname("")
{
}

CSyslogSocket::~CSyslogSocket()
{
}

void CSyslogSocket::Initialize()
{
	try{
		if(AfxSocketInit()==TRUE){
			SourceHostname = DetectHostname();
			BOOL ret = Create(0,SOCK_DGRAM,FD_WRITE);
			if(ret!=TRUE){
				ReportError();
			}else{
				m_IsReady = true;
			}
		}else{
			ReportError();
		}
	}catch(...){
		ReportError();
	}
}

CString CSyslogSocket::DetectHostname()
{
	try{
		TCHAR chrComputerName[MAX_COMPUTERNAME_LENGTH + 1];
		DWORD dwBufferSize = MAX_COMPUTERNAME_LENGTH + 1; 
		if(GetComputerName(chrComputerName,&dwBufferSize)) {
			return chrComputerName;
		} else {
			return "";
		} 
	}catch(...){
		return "";
	}
}

void CSyslogSocket::OnSend(int nErrorCode)
{
	m_IsReady = true;
	CAsyncSocket::OnSend(nErrorCode);
}

bool CSyslogSocket::Send(const void* pBuf, int nBufLen, CString& DestinationHost, UINT DestinationPort)
{
	if(!m_IsReady)
		return false;

	try{
		int ret = CAsyncSocket::SendToEx(pBuf,nBufLen,DestinationPort,DestinationHost);
		if(ret==SOCKET_ERROR){
			UINT uErr = GetLastError();
			if(uErr==WSAEWOULDBLOCK){
				m_IsReady = false;
			}else{
				ReportError(uErr);
			}
			return false;
		}
		return true;
	}catch(...){
		LastError = "Fatal error sending syslog message";
		return false;
	}
}

void CSyslogSocket::ReportError()
{
	try{
		ReportError(GetLastError());
	}catch(...){
		LastError = "Fatal error in call to GetLastError()";
	}
}

void CSyslogSocket::ReportError(UINT uErr)
{
	TCHAR szError[256];
	wsprintf(szError, "Socket Error: %d", uErr);
	LastError = szError;
	//AfxMessageBox(szError);
}