/*
	AQTRONIX WebKnight - ISAPI Filter for securing IIS
	Copyright 2012 Parcifal Aertssen

	This file is part of AQTRONIX WebKnight.

    AQTRONIX WebKnight is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License as
	published by the Free Software Foundation; either version 2 of the
	License, or (at your option) any later version.

    AQTRONIX WebKnight is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with AQTRONIX WebKnight; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include "StdAfx.h"
#include "WebKnightModule.h"

//http://msdn.microsoft.com/en-us/library/ms690697%28v=vs.90%29.aspx
//http://msdn.microsoft.com/en-us/library/ms689348%28v=vs.90%29.aspx
//http://msdn.microsoft.com/en-us/library/ms693073%28v=vs.90%29.aspx

CWebKnightModule::CWebKnightModule(void)
{
	// Open a handle to the Event Log.
    m_hEventLog = RegisterEventSource(NULL,"IISADMIN");
}

CWebKnightModule::~CWebKnightModule(void)
{
	// Test whether the handle for the Event Log is open.
    if (NULL != m_hEventLog){
        // Close the handle to the Event Log.
        DeregisterEventSource(m_hEventLog);
        m_hEventLog = NULL;
    }
}

// Process a GL_APPLICATION_START notification.
GLOBAL_NOTIFICATION_STATUS CWebKnightModule::OnGlobalPreBeginRequest(IN IPreBeginRequestProvider * pProvider)
{
    UNREFERENCED_PARAMETER(pProvider);
	IHttpContext* pc = pProvider->GetHttpContext();
    
	//WriteEventViewerLog("This version of WebKnight does not support IIS Module interface.");

	SendPage(pc, "This version of WebKnight does not support IIS Module interface.");
	return GL_NOTIFICATION_HANDLED;

	
    return GL_NOTIFICATION_CONTINUE;
}

VOID CWebKnightModule::Terminate()
{
    // Remove the class from memory.
    delete this;
}

HRESULT CWebKnightModule::SendPage(IHttpContext* pHttpContext, CString Text)
{
	HRESULT hr = S_FALSE;
	IHttpResponse * pHttpResponse = pHttpContext->GetResponse();
	if(pHttpResponse != NULL){
		HTTP_DATA_CHUNK data;
        pHttpResponse->Clear();
        char* szBuffer = (char *) pHttpContext->AllocateRequestMemory(Text.GetLength()+1);
        data.DataChunkType = HttpDataChunkFromMemory;
        strcpy_s(szBuffer, Text.GetLength(), Text);
        data.FromMemory.pBuffer = (PVOID) szBuffer;
        data.FromMemory.BufferLength = (ULONG) strlen(szBuffer);
        hr = pHttpResponse->WriteEntityChunkByReference(&data, -1);
	}
	return hr;
}

CString CWebKnightModule::GetHeader(IHttpContext* pHttpContext, LPCTSTR Header)
{
    // Buffers to store the returned header value.
    PCSTR pszHeader;

    // Length of the returned header value.
    USHORT cchHeader;

    // Retrieve a pointer to the request.
    IHttpRequest * pHttpRequest = pHttpContext->GetRequest();

    // Test for an error.
    if (pHttpRequest != NULL)
    {
        // Get the lengh of the header.
        pszHeader = pHttpRequest->GetHeader(Header,&cchHeader);

        // The header length will be 0 if the header was not found.
		if (cchHeader == 0 || pszHeader==NULL){
			return "";
		}else{
			return pszHeader;
        }
    }
	return "";
}

BOOL CWebKnightModule::WriteEventLog(LPCSTR szNotification)
{
    // Test whether the handle for the Event Viewer is open.
    if (NULL != m_hEventLog)
    {
        // Write any strings to the Event Viewer and return.
        return ReportEvent(m_hEventLog, EVENTLOG_INFORMATION_TYPE, 0, 0, NULL, 1, 0, &szNotification, NULL);
    }
    return FALSE;
}

// Create the module's exported registration function.
HRESULT __stdcall RegisterModule(DWORD dwServerVersion, IHttpModuleRegistrationInfo * pModuleInfo, IHttpServer * pGlobalInfo)
{
    UNREFERENCED_PARAMETER(dwServerVersion);
    UNREFERENCED_PARAMETER(pGlobalInfo);

    // Create an instance of the global module class.
    CWebKnightModule * pModule = new CWebKnightModule;
    // Test for an error.
    if (NULL == pModule){
        return HRESULT_FROM_WIN32(ERROR_NOT_ENOUGH_MEMORY);
    }
    // Set the global notifications and exit.
    return pModuleInfo->SetGlobalNotifications(pModule, GL_PRE_BEGIN_REQUEST);
}